from .basic.embedding import BasicDocument, BasicEmbeddingPipeline
from .basic.ingestion import BasicIngestionPipeline
from .basic.rag import BasicRAGPipeline
from .web_search.rag import WebSearchRAGPipeline

__all__ = [
    "BasicRAGPipeline",
    "BasicEmbeddingPipeline",
    "BasicIngestionPipeline",
    "BasicDocument",
    "WebSearchRAGPipeline",
]



import uuid

from r2r.client import R2RClient

# Initialize the client with the base URL of your API
base_url = "http://localhost:8000"  # Change this to your actual API base URL
client = R2RClient(base_url)

print("Upserting entry to remote db...")
# Upsert a single entry

entry_response = client.add_entry(
    str(uuid.uuid5(uuid.NAMESPACE_DNS, "doc 1")),  # document_id
    {"txt": "This is a test entry"},
    {"tags": ["example", "test"]},
    do_upsert=True,
)
print(f"Upsert entry response:\n{entry_response}\n\n")


entry_response = client.add_entry(
    str(uuid.uuid5(uuid.NAMESPACE_DNS, "doc 1")),  # document_id
    {"txt": "This is a test entry"},
    {"tags": ["example", "test"]},
    do_upsert=False,
)
print(f"Copy same entry response:\n{entry_response}\n\n")


print("Upserting entries to remote db...")
# Upsert multiple entries
entries = [
    {
        "document_id": str(uuid.uuid5(uuid.NAMESPACE_DNS, "doc 2")),
        "blobs": {"txt": "Second test entry"},
        "metadata": {"tags": "bulk"},
    },
    {
        "document_id": str(uuid.uuid5(uuid.NAMESPACE_DNS, "doc 3")),
        "blobs": {"txt": "Third test entry"},
        "metadata": {"tags": "example"},
    },
]
bulk_upsert_response = client.add_entries(entries, do_upsert=True)
print(f"Upsert entries response:\n{bulk_upsert_response}\n\n")

# Perform a search
print("Searching remote db...")
search_response = client.search("test", 5)
print(f"Search response:\n{search_response}\n\n")

print("Searching remote db with filter...")
# Perform a search w/ filter
filtered_search_response = client.search("test", 5, filters={"tags": "bulk"})
print(f"Search response w/ filter:\n{filtered_search_response}\n\n")

print("Deleting sample document in remote db...")
# Delete a document
response = client.filtered_deletion(
    "document_id", str(uuid.uuid5(uuid.NAMESPACE_DNS, "doc 2"))
)
print(f"Deletion response:\n{response}\n\n")

print("Searching remote db with filter after deletion...")
# Perform a search w/ filter after deletion
post_deletion_filtered_search_response = client.search(
    "test", 5, filters={"tags": "bulk"}
)
print(
    f"Search response w/ filter+deletion:\n{post_deletion_filtered_search_response}\n\n"
)

# Example file path for upload
file_path = "examples/basic/test.pdf"  # Ensure this file exists in your script's directory

print(f"Uploading and processing file: {file_path}...")
# # Upload and process a file
pdf_document_id = str(uuid.uuid5(uuid.NAMESPACE_DNS, "pdf 1"))  # document_id
metadata = {"tags": ["example", "test"]}
upload_pdf_response = client.upload_and_process_file(
    pdf_document_id, file_path, metadata, None
)
print(f"Upload test pdf response:\n{upload_pdf_response}\n\n")

print("Searching remote db after upload...")
# Perform a search on this file
pdf_filtered_search_response = client.search(
    "what is a cool physics equation?",
    5,
    filters={"document_id": pdf_document_id},
)
print(
    f"Search response w/ uploaded pdf filter:\n{pdf_filtered_search_response}\n"
)


print("Performing RAG...")
# Perform a search on this file
pdf_filtered_search_response = client.rag_completion(
    "Are there any test documents?",
    5,
    filters={"document_id": pdf_document_id},
)
print(
    f"Search response w/ uploaded pdf filter:\n{pdf_filtered_search_response}\n"
)


print("Fetching logs after all steps...")
logs_response = client.get_logs()
print(f"Logs response:\n{logs_response}\n")

print("Fetching logs summary after all steps...")
logs_summary_response = client.get_logs_summary()
print(f"Logs summary response:\n{logs_summary_response}\n")



import functools
import os
from typing import Optional


class LoggingDatabaseConnection:
    """
    A class to connect to a database and log the execution of methods to it.
    """

    supported_providers = ["postgres", "local"]

    def __init__(
        self,
        provider="postgres",
        log_table_name="logs",
        local_db_path: Optional[str] = None,
    ):
        if provider not in self.supported_providers:
            raise ValueError(
                f"Error, `{provider}` is not in LoggingDatabaseConnection's list of supported providers."
            )

        self.provider = provider
        self.conn = None
        self.log_table_name = log_table_name

        if self.provider == "postgres":
            if local_db_path is not None:
                raise ValueError(
                    "Error, `local_db_path` should be None when using `postgres`."
                )
            try:
                import psycopg2

                self.db_module = psycopg2
            except ImportError:
                raise ValueError(
                    "Error, `psycopg2` is not installed. Please install it using `pip install psycopg2`."
                )
            if (
                not os.getenv("POSTGRES_DBNAME")
                or not os.getenv("POSTGRES_USER")
                or not os.getenv("POSTGRES_PASSWORD")
                or not os.getenv("POSTGRES_HOST")
                or not os.getenv("POSTGRES_PORT")
            ):
                raise ValueError(
                    "Please set the environment variables POSTGRES_DBNAME, POSTGRES_USER, POSTGRES_PASSWORD, POSTGRES_HOST, and POSTGRES_PORT to run `LoggingDatabaseConnection` with `postgres`."
                )
        elif self.provider == "local":
            import sqlite3

            self.db_module = sqlite3
            self.local_db_path = local_db_path
        else:
            raise ValueError(
                "Invalid provider. Expected 'postgres' or 'local'."
            )

    def __enter__(self):
        if self.provider == "postgres":
            self.conn = self.db_module.connect(
                dbname=os.getenv("POSTGRES_DBNAME"),
                user=os.getenv("POSTGRES_USER"),
                password=os.getenv("POSTGRES_PASSWORD"),
                host=os.getenv("POSTGRES_HOST"),
                port=os.getenv("POSTGRES_PORT"),
            )
            with self.conn.cursor() as cur:
                cur.execute(
                    f"""
                    CREATE TABLE IF NOT EXISTS {self.log_table_name} (
                        timestamp TIMESTAMP,
                        pipeline_run_id UUID,
                        pipeline_run_type TEXT,
                        method TEXT,
                        result TEXT,
                        log_level TEXT
                    )
                """
                )

        elif self.provider == "local":
            self.conn = self.db_module.connect(
                self.local_db_path or os.getenv("LOCAL_DB_PATH")
            )
            self.conn.execute(
                f"""
                CREATE TABLE IF NOT EXISTS {self.log_table_name} (
                    timestamp DATETIME,
                    pipeline_run_id TEXT,
                    pipeline_run_type TEXT,
                    method TEXT,
                    result TEXT,
                    log_level TEXT
                )
                """
            )

        self.conn.commit()
        return self.conn

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.conn:
            self.conn.close()

    def get_logs(self) -> list:
        with self as conn:
            cur = conn.execute(f"SELECT * FROM {self.log_table_name}")
            colnames = [desc[0] for desc in cur.description]
            logs = [dict(zip(colnames, row)) for row in cur.fetchall()]
        return logs


def log_execution_to_db(func):
    """A decorator to log the execution of a method to the database."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # Assuming args[0] is the instance of the class the method belongs to
        instance = args[0]
        inst_provider = instance.logging_provider
        if not inst_provider:
            return func(*args, **kwargs)

        # Adjusted to use 'run_id' and 'type'
        arg_pipeline_run_id = instance.pipeline_run_info["run_id"]
        arg_pipeline_run_type = instance.pipeline_run_info["type"]
        arg_log_table_name = inst_provider.log_table_name

        try:
            # Execute the function and get the result
            result = func(*args, **kwargs)
            log_level = "INFO"
        except Exception as e:
            result = str(e)
            log_level = "ERROR"

        # Log the execution to the database
        timestamp_func = (
            "NOW()"
            if inst_provider.provider == "postgres"
            else "datetime('now')"
        )
        with inst_provider as conn:
            conn.execute(
                f"INSERT INTO {arg_log_table_name} (timestamp, pipeline_run_id, pipeline_run_type, method, result, log_level) VALUES ({timestamp_func}, ?, ?, ?, ?, ?)",
                (
                    str(arg_pipeline_run_id),
                    arg_pipeline_run_type,
                    func.__name__,
                    str(result),
                    log_level,
                ),
            )

            # Commit the transaction
            conn.commit()

        if log_level == "ERROR":
            raise Exception(result)

        return result

    return wrapper



from abc import ABC, abstractmethod
from typing import Any, Generator, Optional


class DatasetConfig:
    def __init__(
        self,
        name: str,
        data_files: Optional[list[str]] = None,
        max_entries: Optional[int] = None,
        text_field: Optional[str] = None,
        split: Optional[str] = "train",
    ):
        """
        Initialize the dataset configuration.

        Args:
            name (str): The name of the dataset.
            data_files (Optional[list[str]]): Specific data files or patterns to load from the dataset.
            max_entries (Optional[int]): Maximum number of entries to stream from the dataset. None for unlimited.
            text_field (Optional[str]): The specific field in the dataset to treat as the main text content.
        """
        self.name = name
        self.data_files = data_files or []
        self.max_entries = max_entries
        self.text_field = text_field
        self.split = split


class DatasetProvider(ABC):
    supported_providers = ["huggingface", "pdf"]

    def __init__(self, provider: str) -> None:
        if provider not in self.supported_providers:
            raise ValueError(
                f"Error, `{provider}` is not in DatasetProvider's list of supported providers."
            )

    @abstractmethod
    def load_datasets(
        self,
        dataset_configs: list[DatasetConfig],
        split: str = "train",
        streaming: bool = True,
    ) -> None:
        """
        Abstract method to load multiple datasets based on their configurations.

        Args:
            dataset_configs (list[DatasetConfig]): Configurations for each dataset to load.
            split (str): The specific split of the dataset to load, e.g., 'train', 'test'.
            streaming (bool): Whether to stream the dataset or load it entirely.
        """
        pass

    @abstractmethod
    def stream_text(self) -> Generator[Any, None, None]:
        """
        Abstract method to stream dataset entries across all configured datasets.

        Yields:
            entry (Dict[str, Any]): A dictionary representing a single dataset entry.
        """
        pass



from typing import Any, Dict, List, Optional, Union

import requests


class R2RClient:
    def __init__(self, base_url: str):
        self.base_url = base_url

    def upload_and_process_file(
        self,
        document_id: str,
        file_path: str,
        metadata: Optional[dict] = None,
        settings: Optional[dict] = None,
    ):
        url = f"{self.base_url}/upload_and_process_file/"
        with open(file_path, "rb") as file:
            import json

            files = {
                "file": (file_path.split("/")[-1], file, "application/pdf")
            }
            data = {
                "document_id": document_id,
                "metadata": json.dumps(metadata)
                if metadata
                else json.dumps({}),
                "settings": json.dumps(settings)
                if settings
                else json.dumps({}),
            }
            response = requests.post(
                url,
                files=files,
                data=data,
            )
        return response.json()

    def add_entry(
        self,
        document_id: str,
        blobs: Dict[str, str],
        metadata: Optional[Dict[str, Any]] = None,
        do_upsert: Optional[bool] = False,
        settings: Optional[Dict[str, Any]] = None,
    ):
        url = f"{self.base_url}/add_entry/"
        json_data = {
            "entry": {
                "document_id": document_id,
                "blobs": blobs,
                "metadata": metadata or {},
            },
            "settings": settings
            or {"embedding_settings": {"do_upsert": do_upsert}},
        }
        response = requests.post(url, json=json_data)
        return response.json()

    def add_entries(
        self,
        entries: List[Dict[str, Any]],
        do_upsert: Optional[bool] = False,
        settings: Optional[Dict[str, Any]] = None,
    ):
        url = f"{self.base_url}/add_entries/"
        json_data = {
            "entries": entries,
            "settings": settings
            or {"embedding_settings": {"do_upsert": do_upsert}},
        }
        response = requests.post(url, json=json_data)
        return response.json()

    def search(
        self,
        query: str,
        limit: Optional[int] = 10,
        filters: Optional[Dict[str, Any]] = None,
        settings: Optional[Dict[str, Any]] = None,
    ):
        url = f"{self.base_url}/search/"
        json_data = {
            "query": query,
            "filters": filters or {},
            "limit": limit,
            "settings": settings or {},
        }
        response = requests.post(url, json=json_data)
        return response.json()

    def rag_completion(
        self,
        query: str,
        limit: Optional[int] = 10,
        filters: Optional[Dict[str, Any]] = None,
        settings: Optional[Dict[str, Any]] = None,
    ):
        url = f"{self.base_url}/rag_completion/"
        json_data = {
            "query": query,
            "filters": filters or {},
            "limit": limit,
            "settings": settings or {},
        }
        response = requests.post(url, json=json_data)
        return response.json()

    def filtered_deletion(self, key: str, value: Union[bool, int, str]):
        url = f"{self.base_url}/filtered_deletion/"
        response = requests.delete(url, params={"key": key, "value": value})
        return response.json()

    def get_logs(self):
        url = f"{self.base_url}/logs"
        response = requests.get(url)
        return response.json()

    def get_logs_summary(self):
        url = f"{self.base_url}/logs_summary"
        response = requests.get(url)
        return response.json()



import json
import logging
import os
import re
from logging.handlers import RotatingFileHandler
from typing import Any

logger = logging.getLogger(__name__)


# Function to find the project root by looking for a .git folder or setup.py file
def find_project_root(current_dir):
    for parent in current_dir.parents:
        if any((parent / marker).exists() for marker in [".git", "setup.py"]):
            return parent
    return current_dir  # Fallback to current dir if no marker found


def load_config(config_path=None):
    if config_path is None:
        # Get the root directory of the project
        root_dir = os.path.dirname(
            os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        )
        config_path = os.path.join(root_dir, "config.json")

    # Load configuration from JSON file
    with open(config_path) as f:
        config = json.load(f)

    # Extract configuration parameters
    api_config = config["api"]
    logging_config = config["logging"]
    embedding_config = config["embedding"]
    database_config = config["database"]
    llm_config = config["language_model"]
    text_splitter_config = config["text_splitter"]

    return (
        api_config,
        logging_config,
        embedding_config,
        database_config,
        llm_config,
        text_splitter_config,
    )


def configure_logging():
    # Determine the root directory of the project
    root_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    logs_dir = os.path.join(
        root_dir, "..", "logs"
    )  # Place the logs directory one level above the root directory

    # Create the logs directory if it doesn't exist
    if not os.path.exists(logs_dir):
        os.makedirs(logs_dir)

    # Create a custom logger
    logger = logging.getLogger("r2r")
    logger.setLevel(logging.DEBUG)  # Set the logging level

    # Create handlers (console and file handler with rotation)
    c_handler = logging.StreamHandler()
    log_file_path = os.path.join(logs_dir, "r2r.log")
    f_handler = RotatingFileHandler(
        log_file_path, maxBytes=1000000, backupCount=5
    )
    c_handler.setLevel(logging.WARNING)  # Console handler level
    f_handler.setLevel(logging.DEBUG)  # File handler level

    # Create formatters and add it to handlers
    c_format = logging.Formatter("%(name)s - %(levelname)s - %(message)s")
    f_format = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    c_handler.setFormatter(c_format)
    f_handler.setFormatter(f_format)

    # Add handlers to the logger
    logger.addHandler(c_handler)
    logger.addHandler(f_handler)


def update_aggregation_entries(
    log: dict[str, Any], event_aggregation: dict[str, dict[str, dict]]
):
    pipeline_run_id = log["pipeline_run_id"]
    if pipeline_run_id is None:
        logger.error(f"Missing 'pipeline_run_id' in log: {log}")
        raise ValueError(f"Missing 'pipeline_run_id' in log: {log}")

    pipeline_run_type = log["pipeline_run_type"]
    if pipeline_run_type is None:
        logger.error(f"Missing 'pipeline_run_type' in log: {log}")
        raise ValueError(f"Missing 'pipeline_run_type' in log: {log}")

    if pipeline_run_id not in event_aggregation:
        event_aggregation[pipeline_run_id] = {
            "timestamp": log["timestamp"],
            "pipeline_run_id": pipeline_run_id,
            "pipeline_run_type": pipeline_run_type,
            "events": [],
        }
    event = {
        "method": log["method"],
        "result": log["result"],
        "log_level": log["log_level"],
        "outcome": "success" if log["log_level"] == "INFO" else "fail",
    }
    event_aggregation[pipeline_run_id]["events"].append(event)


def process_event(event: dict[str, Any]) -> dict[str, Any]:
    method = event["method"]
    result = event.get("result", "N/A")
    processed_result = {}

    if method == "ingress":
        processed_result["search_query"] = result
    elif method == "search":
        text_matches = re.findall(r"'text': '([^']*)'", result)
        scores = re.findall(r"score=(\d+\.\d+)", result)
        processed_result["search_results"] = [
            {"text": text, "score": score}
            for text, score in zip(text_matches, scores)
        ]
        processed_result["method"] = "Search"
    elif method == "generate_completion":
        content_matches = re.findall(r"content='([^']*)'", result)
        processed_result["completion_result"] = ", ".join(content_matches)
        processed_result["method"] = "Generate Completion"

    return processed_result


def combine_aggregated_logs(
    event_aggregation: dict[str, dict[str, Any]]
) -> list[dict[str, Any]]:
    logs_summary = []
    for run_id, aggregation in event_aggregation.items():
        # Assuming 'pipeline_run_type' is available in the log entries to determine the type of pipeline
        pipeline_type = (
            aggregation["pipeline_run_type"]
            if "pipeline_run_type" in aggregation
            else "unknown"
        )

        summary_entry = {
            "timestamp": aggregation["timestamp"],
            "pipeline_run_id": run_id,
            "pipeline_run_type": pipeline_type,
            "method": "",
            "search_query": "",
            "search_results": [],
            # "search_score": "",
            "completion_result": "N/A",  # Default to "N/A" if not applicable
            "outcome": "success"
            if aggregation["events"][-1].get("log_level") == "INFO"
            else "fail",
        }

        for event in aggregation["events"]:
            summary_entry.update(process_event(event))
        logs_summary.append(summary_entry)
    return logs_summary


def process_logs(logs: list[dict[str, Any]]) -> list[dict[str, Any]]:
    event_aggregation = {}
    for log in logs:
        update_aggregation_entries(log, event_aggregation)
    # Convert each aggregated log entry to SummaryLogModel before returning
    return combine_aggregated_logs(event_aggregation)



"""A simple example to demonstrate the usage of `BasicEmbeddingPipeline`."""
import logging
import uuid

import dotenv
from langchain.text_splitter import RecursiveCharacterTextSplitter

from r2r.core import DatasetConfig, LoggingDatabaseConnection
from r2r.datasets import HuggingFaceDataProvider
from r2r.embeddings import OpenAIEmbeddingProvider
from r2r.main import load_config
from r2r.pipelines import BasicDocument, BasicEmbeddingPipeline
from r2r.vector_dbs import PGVectorDB, QdrantDB

if __name__ == "__main__":
    dotenv.load_dotenv()

    (
        api_config,
        logging_config,
        embedding_config,
        database_config,
        llm_config,
        text_splitter_config,
    ) = load_config()

    logger = logging.getLogger(logging_config["name"])
    logging.basicConfig(level=logging_config["level"])

    logger.info("Starting the embedding pipeline")

    # Specify the embedding provider
    embeddings_provider = OpenAIEmbeddingProvider()
    embedding_model = embedding_config["model"]
    embedding_dimension = embedding_config["dimension"]
    embedding_batch_size = embedding_config["batch_size"]

    # Specify the vector database provider
    db = (
        QdrantDB() if database_config["provider"] == "qdrant" else PGVectorDB()
    )
    collection_name = database_config["collection_name"]
    db.initialize_collection(collection_name, embedding_dimension)

    # Specify the dataset providers
    dataset_provider = HuggingFaceDataProvider()
    dataset_provider.load_datasets(
        [
            DatasetConfig("camel-ai/physics", None, 10, "message_2"),
            DatasetConfig("camel-ai/chemistry", None, 10, "message_2"),
        ],
    )

    # Specify the chunking strategy
    text_splitter = RecursiveCharacterTextSplitter(
        chunk_size=text_splitter_config["chunk_size"],
        chunk_overlap=text_splitter_config["chunk_overlap"],
        length_function=len,
        is_separator_regex=False,
    )

    logging_provider = LoggingDatabaseConnection(logging_config["database"])

    pipeline = BasicEmbeddingPipeline(
        embedding_model,
        embeddings_provider,
        db,
        text_splitter=text_splitter,
        embedding_batch_size=embedding_batch_size,
        logging_provider=logging_provider,
    )

    entry_id = 0
    document_batch = []
    for entry in dataset_provider.stream_text():
        if entry is None:
            break
        text, config = entry
        document_id = str(uuid.uuid5(uuid.NAMESPACE_URL, config.name))

        if text is None:
            break
        document_batch.append(
            BasicDocument(
                id=str(
                    uuid.uuid5(uuid.NAMESPACE_URL, f"{config.name}_{text}")
                ),
                text=text,
                metadata={"document_id": document_id},
            )
        )
        entry_id += 1

        if len(document_batch) == 1:
            logging.info(
                f"Processing batch of {len(document_batch)} documents."
            )
            pipeline.run(document_batch)
            document_batch = []

    logging.info(f"Processing final {len(document_batch)} documents.")
    pipeline.run(document_batch)

    pipeline.close()



import json
import logging
from datetime import datetime
from pathlib import Path
from typing import Optional, Union

from fastapi import FastAPI, File, Form, HTTPException, UploadFile
from fastapi.middleware.cors import CORSMiddleware
from pydantic import BaseModel, Field

from r2r.core import (
    EmbeddingPipeline,
    IngestionPipeline,
    LoggingDatabaseConnection,
    RAGPipeline,
)
from r2r.main.utils import configure_logging, find_project_root, process_logs

logger = logging.getLogger("r2r")

# Current directory where this script is located
CURRENT_DIR = Path(__file__).resolve().parent


def create_app(
    ingestion_pipeline: IngestionPipeline,
    embedding_pipeline: EmbeddingPipeline,
    rag_pipeline: RAGPipeline,
    upload_path: Optional[Path] = None,
    logging_provider: Optional[LoggingDatabaseConnection] = None,
):
    app = FastAPI()
    configure_logging()

    # CORS setup
    origins = [
        "*",  # TODO - Change this to the actual frontend URL
        "http://localhost:3000",
        "http://localhost:8000",
    ]

    app.add_middleware(
        CORSMiddleware,
        allow_origins=origins,  # Allows specified origins
        allow_credentials=True,
        allow_methods=["*"],  # Allows all methods
        allow_headers=["*"],  # Allows all headers
    )

    upload_path = upload_path or find_project_root(CURRENT_DIR) / "uploads"

    if not upload_path.exists():
        upload_path.mkdir()

    class EmbeddingsSettingsModel(BaseModel):
        do_chunking: Optional[bool] = True
        do_upsert: Optional[bool] = True

    class IngestionSettingsModel(BaseModel):
        pass

    class RAGSettingsModel(BaseModel):
        pass

    class SettingsModel(BaseModel):
        embedding_settings: EmbeddingsSettingsModel = EmbeddingsSettingsModel()
        ingestion_settings: IngestionSettingsModel = IngestionSettingsModel()
        rag_settings: RAGSettingsModel = RAGSettingsModel()

    class EntryModel(BaseModel):
        document_id: str
        blobs: dict[str, str]
        metadata: Optional[dict]

    # File upload class, to be fixed later
    # class FileUploadRequest(BaseModel):
    # document_id: str
    # metadata: Optional[dict]
    # settings: SettingsModel = SettingsModel()

    class AddEntryRequest(BaseModel):
        entry: EntryModel
        settings: SettingsModel = SettingsModel()

    class AddEntriesRequest(BaseModel):
        entries: list[EntryModel]
        settings: SettingsModel = SettingsModel()

    class RAGQueryModel(BaseModel):
        query: str
        limit: Optional[int] = 10
        filters: dict = {}
        settings: SettingsModel = SettingsModel()

    def to_camel(string: str) -> str:
        return "".join(
            word.capitalize() if i != 0 else word
            for i, word in enumerate(string.split("_"))
        )

    class LogModel(BaseModel):
        timestamp: datetime = Field(alias="timestamp")
        pipeline_run_id: str = Field(alias="pipelineRunId")
        pipeline_run_type: str = Field(alias="pipelineRunType")
        method: str = Field(alias="method")
        result: str = Field(alias="result")
        log_level: str = Field(alias="logLevel")

        class Config:
            alias_generator = to_camel
            allow_population_by_field_name = True

    class SummaryLogModel(BaseModel):
        timestamp: datetime = Field(alias="timestamp")
        pipeline_run_id: str = Field(alias="pipelineRunId")
        pipeline_run_type: str = Field(alias="pipelineRunType")
        method: str = Field(alias="method")
        search_query: str = Field(alias="searchQuery")
        search_results: list[dict] = Field(alias="searchResults")
        completion_result: str = Field(alias="completionResult")
        outcome: str = Field(alias="outcome")

        class Config:
            alias_generator = to_camel
            allow_population_by_field_name = True

    @app.post("/upload_and_process_file/")
    # TODO - Why can't we use a BaseModel to represent the request?
    # Naive class FileUploadRequest(BaseModel) above fails
    async def upload_and_process_file(
        document_id: str = Form(...),
        metadata: str = Form("{}"),
        settings: str = Form("{}"),
        file: UploadFile = File(...),
    ):
        metadata_json = json.loads(metadata)
        settings_model = SettingsModel.parse_raw(settings)

        if not file.filename:
            raise HTTPException(
                status_code=400, detail="No file was uploaded."
            )
        # Extract file extension and check if it's an allowed type
        file_extension = file.filename.split(".")[-1]
        if file_extension not in ingestion_pipeline.supported_types:
            raise HTTPException(
                status_code=400,
                detail=f"Invalid file type. Allowed types are: {', '.join(ingestion_pipeline.supported_types)}.",
            )

        file_location = upload_path / file.filename
        try:
            file_content = file.file.read()

            # TODO - Consider saving file to disk
            # with open(file_location, "wb+") as file_object:
            # file_object.write(file_content)

            document = ingestion_pipeline.run(
                document_id,
                {file_extension: file_content},
                metadata=metadata_json,
                **settings_model.ingestion_settings.dict(),
            )
            embedding_pipeline.run(
                document, **settings_model.embedding_settings.dict()
            )

            return {
                "message": f"File '{file.filename}' processed and saved at '{file_location}'"
            }
        except Exception as e:
            logger.error(
                f"upload_and_process_file: [Error](file={file.filename}, error={str(e)})"
            )
            raise HTTPException(status_code=500, detail=str(e))

    @app.post("/add_entry/")
    def add_entry(entry_req: AddEntryRequest):
        try:
            document = ingestion_pipeline.run(
                entry_req.entry.document_id,
                entry_req.entry.blobs,
                metadata=entry_req.entry.metadata,
                **entry_req.settings.ingestion_settings.dict(),
            )
            embedding_pipeline.run(
                document, **entry_req.settings.embedding_settings.dict()
            )
            return {"message": "Entry upserted successfully."}
        except Exception as e:
            logger.error(
                f":add_entry: [Error](entry={entry_req}, error={str(e)})"
            )
            raise HTTPException(status_code=500, detail=str(e))

    @app.post("/add_entries/")
    def add_entries(entries_req: AddEntriesRequest):
        try:
            for entry in entries_req.entries:
                documents = ingestion_pipeline.run(
                    entry.document_id,
                    entry.blobs,
                    metadata=entry.metadata,
                    **entries_req.settings.ingestion_settings.dict(),
                )
                embedding_pipeline.run(
                    documents, **entries_req.settings.embedding_settings.dict()
                )
            return {"message": "Entries upserted successfully."}
        except Exception as e:
            logger.error(
                f":add_entries: [Error](entries={entries_req}, error={str(e)})"
            )
            raise HTTPException(status_code=500, detail=str(e))

    @app.post("/search/")
    def search(query: RAGQueryModel):
        try:
            completion = rag_pipeline.run(
                query.query, query.filters, query.limit, search_only=True
            )
            return completion
        except Exception as e:
            logger.error(f":search: [Error](query={query}, error={str(e)})")
            raise HTTPException(status_code=500, detail=str(e))

    @app.post("/rag_completion/")
    def rag_completion(query: RAGQueryModel):
        try:
            completion = rag_pipeline.run(
                query.query, query.filters, query.limit
            )
            return completion
        except Exception as e:
            logger.error(
                f":rag_completion: [Error](query={query}, error={str(e)})"
            )
            raise HTTPException(status_code=500, detail=str(e))

    @app.delete("/filtered_deletion/")
    def filtered_deletion(key: str, value: Union[bool, int, str]):
        try:
            embedding_pipeline.db.filtered_deletion(key, value)
            return {"message": "Entries deleted successfully."}
        except Exception as e:
            logger.error(
                f":filtered_deletion: [Error](key={key}, value={value}, error={str(e)})"
            )
            raise HTTPException(status_code=500, detail=str(e))

    @app.get("/logs")
    def logs():
        try:
            if logging_provider is None:
                raise HTTPException(
                    status_code=404, detail="Logging provider not found."
                )
            logs = logging_provider.get_logs()
            return {
                "logs": [LogModel(**log).dict(by_alias=True) for log in logs]
            }
        except Exception as e:
            logger.error(f":logs: [Error](error={str(e)})")
            raise HTTPException(status_code=500, detail=str(e))

    @app.get("/logs_summary")
    def logs_summary():
        try:
            if logging_provider is None:
                raise HTTPException(
                    status_code=404, detail="Logging provider not found."
                )
            logs = logging_provider.get_logs()
            logs_summary = process_logs(logs)
            events_summary = [
                SummaryLogModel(**log).dict(by_alias=True)
                for log in logs_summary
            ]
            return {"events_summary": events_summary}
        except Exception as e:
            logger.error(f":logs_summary: [Error](error={str(e)})")
            raise HTTPException(status_code=500, detail=str(e))

    return app



import os

from r2r.codesearch.scip_pb2 import Index


class Indexer:
    def __init__(self, path: str = "index.scip", project_name: str = "r2r"):
        self.path = path
        self.project_name = project_name
        self.index = self._load_index_protobuf()

    def _load_index_protobuf(self) -> Index:
        """
        Loads and returns an Index protobuf object from the given file path.
        """
        index = Index()
        with open(self.path, "rb") as f:
            index.ParseFromString(f.read())
        return index

    def extractor(self):
        for document in self.index.documents:
            with open(
                os.path.join(self.project_name, document.relative_path), "r"
            ) as f:
                content = f.readlines()
            for i, occurence in enumerate(document.occurrences):
                if "python r2r" not in occurence.symbol:
                    continue
                if len(occurence.enclosing_range) != 4:
                    continue
                # bounding box coordinates
                symbol = "".join(occurence.symbol.split("`")[1:])
                line, col, end_line, end_col = occurence.enclosing_range
                extraction = "".join(content[line:end_line])
                yield symbol, extraction



import logging
from typing import Generator, Optional, Tuple

from r2r.core import DatasetConfig, DatasetProvider

logger = logging.getLogger(__name__)


class HuggingFaceDataProvider(DatasetProvider):
    def __init__(self, provider: str = "huggingface"):
        if provider != "huggingface":
            raise ValueError(
                f"Error, HuggingFaceDataProvider was passed `{provider}` as a provider."
            )
        super().__init__(provider)
        try:
            import datasets  # noqa
        except ImportError:
            raise ValueError(
                f"Error, `datasets` must be installed to create a `HuggingFaceDataProvider` object. Please install it using `pip install datasets`."
            )

    def load_datasets(
        self,
        dataset_configs: list[DatasetConfig],
        split: str = "train",
        streaming: bool = True,
    ) -> None:
        logger.info("Loading datasets with HuggingFaceDataProvider now.")
        from datasets import load_dataset

        self.datasets = []  # Prepare to store loaded datasets
        for config in dataset_configs:
            logger.info(
                f"Loading dataset {config.name} with text field {config.text_field} and max entries {config.max_entries}."
            )
            dataset = load_dataset(
                config.name, split=config.split, streaming=streaming
            )
            self.datasets.append((dataset, config))

    def stream_text(
        self,
    ) -> Generator[Optional[Tuple[str, DatasetConfig]], None, None]:
        logger.info("Streaming text with with HuggingFaceDataProvider now.")
        for dataset, config in self.datasets:
            entries_streamed = 0
            for entry in dataset:
                if (
                    config.max_entries is not None
                    and entries_streamed >= config.max_entries
                ):
                    break  # Stop streaming if max_entries is reached

                # Extract text using the specified text field or default to 'text' field
                text_field = config.text_field if config.text_field else "text"
                text = entry[text_field] if text_field in entry else None

                if text:
                    yield text, config
                    entries_streamed += 1



"""A module for creating OpenAI model abstractions."""
import logging
import os
from dataclasses import dataclass

from openai.types import Completion
from openai.types.chat import ChatCompletion

from r2r.core import GenerationConfig, LLMConfig, LLMProvider

logger = logging.getLogger(__name__)


@dataclass
class OpenAIConfig(LLMConfig):
    """Configuration for OpenAI models."""

    # Base
    provider_name: str = "openai"


class OpenAILLM(LLMProvider):
    """A concrete class for creating OpenAI models."""

    PROMPT_MEASUREMENT_PREFIX = (
        "<DUMMY PADDING, LOOKUP ACTUAL STRUCTUER LATER>"
    )

    def __init__(
        self,
        config: OpenAIConfig,
        *args,
        **kwargs,
    ) -> None:
        logger.info(f"Initializing `OpenAILLM` with config: {config}")
        super().__init__()
        self.config: OpenAIConfig = config

        try:
            from openai import OpenAI  # noqa
        except ImportError:
            raise ImportError(
                "Please install the openai package before attempting to run with an OpenAI model. This can be accomplished via `pip install openai`."
            )
        if config.provider_name != "openai" or not os.getenv("OPENAI_API_KEY"):
            raise ValueError(
                "OpenAI API key not found. Please set the OPENAI_API_KEY environment variable."
            )
        # set the config here, again, for typing purposes
        if not isinstance(self.config, OpenAIConfig):
            raise ValueError(
                "The provided config must be an instance of OpenAIConfig."
            )
        self.client = OpenAI()

    def get_chat_completion(
        self,
        messages: list[dict],
        generation_config: GenerationConfig,
        **kwargs,
    ) -> ChatCompletion:
        """Get a completion from the OpenAI API based on the provided messages."""

        # Create a dictionary with the default arguments
        args = self._get_base_args(
            generation_config,
            OpenAILLM.PROMPT_MEASUREMENT_PREFIX
            + f"{OpenAILLM.PROMPT_MEASUREMENT_PREFIX}\n\n".join(
                [m["content"] for m in messages]
            ),
        )

        args["messages"] = messages

        # Conditionally add the 'functions' argument if it's not None
        if generation_config.functions is not None:
            args["functions"] = generation_config.functions

        args = {**args, **kwargs}
        # Create the chat completion
        return self.client.chat.completions.create(**args)

    def get_instruct_completion(
        self,
        prompt: str,
        generation_config: GenerationConfig,
        **kwargs,
    ) -> Completion:
        """Get an instruction completion from the OpenAI API based on the provided prompt."""

        args = self._get_base_args(generation_config, prompt)

        args["prompt"] = prompt

        # Create the instruction completion
        return self.client.completions.create(**args)

    def _get_base_args(
        self,
        generation_config: GenerationConfig,
        prompt=None,
    ) -> dict:
        """Get the base arguments for the OpenAI API."""

        args = {
            "model": generation_config.model_name,
            "temperature": generation_config.temperature,
            "top_p": generation_config.top_p,
            "stream": generation_config.do_stream,
            # TODO - We need to cap this to avoid potential errors when exceed max allowable context
            "max_tokens": generation_config.max_tokens_to_sample,
        }

        return args



import uuid
from abc import abstractmethod
from typing import Any, Optional

from ..abstractions.document import BasicDocument
from ..providers.logging import LoggingDatabaseConnection
from .pipeline import Pipeline


class IngestionPipeline(Pipeline):
    def __init__(
        self,
        logging_provider: Optional[LoggingDatabaseConnection] = None,
        **kwargs,
    ):
        super().__init__(logging_provider=logging_provider, **kwargs)

    def initialize_pipeline(self) -> None:
        self.pipeline_run_info = {"run_id": uuid.uuid4(), "type": "ingestion"}

    @property
    @abstractmethod
    def supported_types(self) -> list[str]:
        """
        Returns a list of supported data types.
        """
        pass

    @abstractmethod
    def process_data(self, entry_type: str, entry_data: Any) -> str:
        """
        Process data into plaintext based on the data type.
        """
        pass

    @abstractmethod
    def parse_entry(self, entry_type: str, entry_data: Any) -> str:
        """
        Parse entry data into plaintext based on the entry type.
        """
        pass

    def run(
        self,
        document_id: str,
        blobs: dict[str, Any],
        metadata: Optional[dict] = None,
        **kwargs,
    ) -> BasicDocument:
        """
        Run the appropriate parsing method based on the data type and whether the data is a file or an entry.
        Returns the processed data and metadata.
        """

        self.initialize_pipeline()

        if len(blobs) == 0:
            raise ValueError("No blobs provided to process.")

        processed_text = ""
        for entry_type, blob in blobs.items():
            if entry_type not in self.supported_types:
                raise ValueError(f"EntryType {entry_type} not supported.")
            processed_text += self.parse_entry(entry_type, blob)

        return BasicDocument(
            id=document_id, text=processed_text, metadata=metadata
        )



# type: ignore
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: scip.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import enum_type_wrapper

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\nscip.proto\x12\x04scip"\x7f\n\x05Index\x12 \n\x08metadata\x18\x01 \x01(\x0b\x32\x0e.scip.Metadata\x12!\n\tdocuments\x18\x02 \x03(\x0b\x32\x0e.scip.Document\x12\x31\n\x10\x65xternal_symbols\x18\x03 \x03(\x0b\x32\x17.scip.SymbolInformation"\x9f\x01\n\x08Metadata\x12&\n\x07version\x18\x01 \x01(\x0e\x32\x15.scip.ProtocolVersion\x12!\n\ttool_info\x18\x02 \x01(\x0b\x32\x0e.scip.ToolInfo\x12\x14\n\x0cproject_root\x18\x03 \x01(\t\x12\x32\n\x16text_document_encoding\x18\x04 \x01(\x0e\x32\x12.scip.TextEncoding"<\n\x08ToolInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\x12\x11\n\targuments\x18\x03 \x03(\t"\x84\x01\n\x08\x44ocument\x12\x10\n\x08language\x18\x04 \x01(\t\x12\x15\n\rrelative_path\x18\x01 \x01(\t\x12%\n\x0boccurrences\x18\x02 \x03(\x0b\x32\x10.scip.Occurrence\x12(\n\x07symbols\x18\x03 \x03(\x0b\x32\x17.scip.SymbolInformation"_\n\x06Symbol\x12\x0e\n\x06scheme\x18\x01 \x01(\t\x12\x1e\n\x07package\x18\x02 \x01(\x0b\x32\r.scip.Package\x12%\n\x0b\x64\x65scriptors\x18\x03 \x03(\x0b\x32\x10.scip.Descriptor"9\n\x07Package\x12\x0f\n\x07manager\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0f\n\x07version\x18\x03 \x01(\t"\x82\x02\n\nDescriptor\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x15\n\rdisambiguator\x18\x02 \x01(\t\x12\'\n\x06suffix\x18\x03 \x01(\x0e\x32\x17.scip.Descriptor.Suffix"\xa5\x01\n\x06Suffix\x12\x15\n\x11UnspecifiedSuffix\x10\x00\x12\r\n\tNamespace\x10\x01\x12\x0f\n\x07Package\x10\x01\x1a\x02\x08\x01\x12\x08\n\x04Type\x10\x02\x12\x08\n\x04Term\x10\x03\x12\n\n\x06Method\x10\x04\x12\x11\n\rTypeParameter\x10\x05\x12\r\n\tParameter\x10\x06\x12\x08\n\x04Meta\x10\x07\x12\t\n\x05Local\x10\x08\x12\t\n\x05Macro\x10\t\x1a\x02\x10\x01"e\n\x11SymbolInformation\x12\x0e\n\x06symbol\x18\x01 \x01(\t\x12\x15\n\rdocumentation\x18\x03 \x03(\t\x12)\n\rrelationships\x18\x04 \x03(\x0b\x32\x12.scip.Relationship"\x82\x01\n\x0cRelationship\x12\x0e\n\x06symbol\x18\x01 \x01(\t\x12\x14\n\x0cis_reference\x18\x02 \x01(\x08\x12\x19\n\x11is_implementation\x18\x03 \x01(\x08\x12\x1a\n\x12is_type_definition\x18\x04 \x01(\x08\x12\x15\n\ris_definition\x18\x05 \x01(\x08"\xc8\x01\n\nOccurrence\x12\r\n\x05range\x18\x01 \x03(\x05\x12\x0e\n\x06symbol\x18\x02 \x01(\t\x12\x14\n\x0csymbol_roles\x18\x03 \x01(\x05\x12\x1e\n\x16override_documentation\x18\x04 \x03(\t\x12%\n\x0bsyntax_kind\x18\x05 \x01(\x0e\x32\x10.scip.SyntaxKind\x12%\n\x0b\x64iagnostics\x18\x06 \x03(\x0b\x32\x10.scip.Diagnostic\x12\x17\n\x0f\x65nclosing_range\x18\x07 \x03(\x05"\x80\x01\n\nDiagnostic\x12 \n\x08severity\x18\x01 \x01(\x0e\x32\x0e.scip.Severity\x12\x0c\n\x04\x63ode\x18\x02 \x01(\t\x12\x0f\n\x07message\x18\x03 \x01(\t\x12\x0e\n\x06source\x18\x04 \x01(\t\x12!\n\x04tags\x18\x05 \x03(\x0e\x32\x13.scip.DiagnosticTag*1\n\x0fProtocolVersion\x12\x1e\n\x1aUnspecifiedProtocolVersion\x10\x00*@\n\x0cTextEncoding\x12\x1b\n\x17UnspecifiedTextEncoding\x10\x00\x12\x08\n\x04UTF8\x10\x01\x12\t\n\x05UTF16\x10\x02*}\n\nSymbolRole\x12\x19\n\x15UnspecifiedSymbolRole\x10\x00\x12\x0e\n\nDefinition\x10\x01\x12\n\n\x06Import\x10\x02\x12\x0f\n\x0bWriteAccess\x10\x04\x12\x0e\n\nReadAccess\x10\x08\x12\r\n\tGenerated\x10\x10\x12\x08\n\x04Test\x10 *\xea\x06\n\nSyntaxKind\x12\x19\n\x15UnspecifiedSyntaxKind\x10\x00\x12\x0b\n\x07\x43omment\x10\x01\x12\x18\n\x14PunctuationDelimiter\x10\x02\x12\x16\n\x12PunctuationBracket\x10\x03\x12\x0b\n\x07Keyword\x10\x04\x12\x19\n\x11IdentifierKeyword\x10\x04\x1a\x02\x08\x01\x12\x16\n\x12IdentifierOperator\x10\x05\x12\x0e\n\nIdentifier\x10\x06\x12\x15\n\x11IdentifierBuiltin\x10\x07\x12\x12\n\x0eIdentifierNull\x10\x08\x12\x16\n\x12IdentifierConstant\x10\t\x12\x1b\n\x17IdentifierMutableGlobal\x10\n\x12\x17\n\x13IdentifierParameter\x10\x0b\x12\x13\n\x0fIdentifierLocal\x10\x0c\x12\x16\n\x12IdentifierShadowed\x10\r\x12\x17\n\x13IdentifierNamespace\x10\x0e\x12\x18\n\x10IdentifierModule\x10\x0e\x1a\x02\x08\x01\x12\x16\n\x12IdentifierFunction\x10\x0f\x12 \n\x1cIdentifierFunctionDefinition\x10\x10\x12\x13\n\x0fIdentifierMacro\x10\x11\x12\x1d\n\x19IdentifierMacroDefinition\x10\x12\x12\x12\n\x0eIdentifierType\x10\x13\x12\x19\n\x15IdentifierBuiltinType\x10\x14\x12\x17\n\x13IdentifierAttribute\x10\x15\x12\x0f\n\x0bRegexEscape\x10\x16\x12\x11\n\rRegexRepeated\x10\x17\x12\x11\n\rRegexWildcard\x10\x18\x12\x12\n\x0eRegexDelimiter\x10\x19\x12\r\n\tRegexJoin\x10\x1a\x12\x11\n\rStringLiteral\x10\x1b\x12\x17\n\x13StringLiteralEscape\x10\x1c\x12\x18\n\x14StringLiteralSpecial\x10\x1d\x12\x14\n\x10StringLiteralKey\x10\x1e\x12\x14\n\x10\x43haracterLiteral\x10\x1f\x12\x12\n\x0eNumericLiteral\x10 \x12\x12\n\x0e\x42ooleanLiteral\x10!\x12\x07\n\x03Tag\x10"\x12\x10\n\x0cTagAttribute\x10#\x12\x10\n\x0cTagDelimiter\x10$\x1a\x02\x10\x01*V\n\x08Severity\x12\x17\n\x13UnspecifiedSeverity\x10\x00\x12\t\n\x05\x45rror\x10\x01\x12\x0b\n\x07Warning\x10\x02\x12\x0f\n\x0bInformation\x10\x03\x12\x08\n\x04Hint\x10\x04*N\n\rDiagnosticTag\x12\x1c\n\x18UnspecifiedDiagnosticTag\x10\x00\x12\x0f\n\x0bUnnecessary\x10\x01\x12\x0e\n\nDeprecated\x10\x02*\xe0\x08\n\x08Language\x12\x17\n\x13UnspecifiedLanguage\x10\x00\x12\x08\n\x04\x41\x42\x41P\x10<\x12\x07\n\x03\x41PL\x10\x31\x12\x07\n\x03\x41\x64\x61\x10\'\x12\x08\n\x04\x41gda\x10-\x12\x0c\n\x08\x41sciiDoc\x10V\x12\x0c\n\x08\x41ssembly\x10:\x12\x07\n\x03\x41wk\x10\x42\x12\x07\n\x03\x42\x61t\x10\x44\x12\n\n\x06\x42ibTeX\x10Q\x12\x05\n\x01\x43\x10"\x12\t\n\x05\x43OBOL\x10;\x12\x07\n\x03\x43PP\x10#\x12\x07\n\x03\x43SS\x10\x1a\x12\n\n\x06\x43Sharp\x10\x01\x12\x0b\n\x07\x43lojure\x10\x08\x12\x10\n\x0c\x43offeescript\x10\x15\x12\x0e\n\nCommonLisp\x10\t\x12\x07\n\x03\x43oq\x10/\x12\x08\n\x04\x44\x61rt\x10\x03\x12\n\n\x06\x44\x65lphi\x10\x39\x12\x08\n\x04\x44iff\x10X\x12\x0e\n\nDockerfile\x10P\x12\n\n\x06\x44yalog\x10\x32\x12\n\n\x06\x45lixir\x10\x11\x12\n\n\x06\x45rlang\x10\x12\x12\n\n\x06\x46Sharp\x10*\x12\x08\n\x04\x46ish\x10\x41\x12\x08\n\x04\x46low\x10\x18\x12\x0b\n\x07\x46ortran\x10\x38\x12\x0e\n\nGit_Commit\x10[\x12\x0e\n\nGit_Config\x10Y\x12\x0e\n\nGit_Rebase\x10\\\x12\x06\n\x02Go\x10!\x12\n\n\x06Groovy\x10\x07\x12\x08\n\x04HTML\x10\x1e\x12\x08\n\x04Hack\x10\x14\x12\x0e\n\nHandlebars\x10Z\x12\x0b\n\x07Haskell\x10,\x12\t\n\x05Idris\x10.\x12\x07\n\x03Ini\x10H\x12\x05\n\x01J\x10\x33\x12\x08\n\x04JSON\x10K\x12\x08\n\x04Java\x10\x06\x12\x0e\n\nJavaScript\x10\x16\x12\x13\n\x0fJavaScriptReact\x10]\x12\x0b\n\x07Jsonnet\x10L\x12\t\n\x05Julia\x10\x37\x12\n\n\x06Kotlin\x10\x04\x12\t\n\x05LaTeX\x10S\x12\x08\n\x04Lean\x10\x30\x12\x08\n\x04Less\x10\x1b\x12\x07\n\x03Lua\x10\x0c\x12\x0c\n\x08Makefile\x10O\x12\x0c\n\x08Markdown\x10T\x12\n\n\x06Matlab\x10\x34\x12\x07\n\x03Nix\x10M\x12\t\n\x05OCaml\x10)\x12\x0f\n\x0bObjective_C\x10$\x12\x11\n\rObjective_CPP\x10%\x12\x07\n\x03PHP\x10\x13\x12\t\n\x05PLSQL\x10\x46\x12\x08\n\x04Perl\x10\r\x12\x0e\n\nPowerShell\x10\x43\x12\n\n\x06Prolog\x10G\x12\n\n\x06Python\x10\x0f\x12\x05\n\x01R\x10\x36\x12\n\n\x06Racket\x10\x0b\x12\x08\n\x04Raku\x10\x0e\x12\t\n\x05Razor\x10>\x12\x08\n\x04ReST\x10U\x12\x08\n\x04Ruby\x10\x10\x12\x08\n\x04Rust\x10(\x12\x07\n\x03SAS\x10=\x12\x08\n\x04SCSS\x10\x1d\x12\x07\n\x03SML\x10+\x12\x07\n\x03SQL\x10\x45\x12\x08\n\x04Sass\x10\x1c\x12\t\n\x05Scala\x10\x05\x12\n\n\x06Scheme\x10\n\x12\x0f\n\x0bShellScript\x10@\x12\x0b\n\x07Skylark\x10N\x12\t\n\x05Swift\x10\x02\x12\x08\n\x04TOML\x10I\x12\x07\n\x03TeX\x10R\x12\x0e\n\nTypeScript\x10\x17\x12\x13\n\x0fTypeScriptReact\x10^\x12\x0f\n\x0bVisualBasic\x10?\x12\x07\n\x03Vue\x10\x19\x12\x0b\n\x07Wolfram\x10\x35\x12\x07\n\x03XML\x10\x1f\x12\x07\n\x03XSL\x10 \x12\x08\n\x04YAML\x10J\x12\x07\n\x03Zig\x10&B/Z-github.com/sourcegraph/scip/bindings/go/scip/b\x06proto3'
)

_PROTOCOLVERSION = DESCRIPTOR.enum_types_by_name["ProtocolVersion"]
ProtocolVersion = enum_type_wrapper.EnumTypeWrapper(_PROTOCOLVERSION)
_TEXTENCODING = DESCRIPTOR.enum_types_by_name["TextEncoding"]
TextEncoding = enum_type_wrapper.EnumTypeWrapper(_TEXTENCODING)
_SYMBOLROLE = DESCRIPTOR.enum_types_by_name["SymbolRole"]
SymbolRole = enum_type_wrapper.EnumTypeWrapper(_SYMBOLROLE)
_SYNTAXKIND = DESCRIPTOR.enum_types_by_name["SyntaxKind"]
SyntaxKind = enum_type_wrapper.EnumTypeWrapper(_SYNTAXKIND)
_SEVERITY = DESCRIPTOR.enum_types_by_name["Severity"]
Severity = enum_type_wrapper.EnumTypeWrapper(_SEVERITY)
_DIAGNOSTICTAG = DESCRIPTOR.enum_types_by_name["DiagnosticTag"]
DiagnosticTag = enum_type_wrapper.EnumTypeWrapper(_DIAGNOSTICTAG)
_LANGUAGE = DESCRIPTOR.enum_types_by_name["Language"]
Language = enum_type_wrapper.EnumTypeWrapper(_LANGUAGE)
UnspecifiedProtocolVersion = 0
UnspecifiedTextEncoding = 0
UTF8 = 1
UTF16 = 2
UnspecifiedSymbolRole = 0
Definition = 1
Import = 2
WriteAccess = 4
ReadAccess = 8
Generated = 16
Test = 32
UnspecifiedSyntaxKind = 0
Comment = 1
PunctuationDelimiter = 2
PunctuationBracket = 3
Keyword = 4
IdentifierKeyword = 4
IdentifierOperator = 5
Identifier = 6
IdentifierBuiltin = 7
IdentifierNull = 8
IdentifierConstant = 9
IdentifierMutableGlobal = 10
IdentifierParameter = 11
IdentifierLocal = 12
IdentifierShadowed = 13
IdentifierNamespace = 14
IdentifierModule = 14
IdentifierFunction = 15
IdentifierFunctionDefinition = 16
IdentifierMacro = 17
IdentifierMacroDefinition = 18
IdentifierType = 19
IdentifierBuiltinType = 20
IdentifierAttribute = 21
RegexEscape = 22
RegexRepeated = 23
RegexWildcard = 24
RegexDelimiter = 25
RegexJoin = 26
StringLiteral = 27
StringLiteralEscape = 28
StringLiteralSpecial = 29
StringLiteralKey = 30
CharacterLiteral = 31
NumericLiteral = 32
BooleanLiteral = 33
Tag = 34
TagAttribute = 35
TagDelimiter = 36
UnspecifiedSeverity = 0
Error = 1
Warning = 2
Information = 3
Hint = 4
UnspecifiedDiagnosticTag = 0
Unnecessary = 1
Deprecated = 2
UnspecifiedLanguage = 0
ABAP = 60
APL = 49
Ada = 39
Agda = 45
AsciiDoc = 86
Assembly = 58
Awk = 66
Bat = 68
BibTeX = 81
C = 34
COBOL = 59
CPP = 35
CSS = 26
CSharp = 1
Clojure = 8
Coffeescript = 21
CommonLisp = 9
Coq = 47
Dart = 3
Delphi = 57
Diff = 88
Dockerfile = 80
Dyalog = 50
Elixir = 17
Erlang = 18
FSharp = 42
Fish = 65
Flow = 24
Fortran = 56
Git_Commit = 91
Git_Config = 89
Git_Rebase = 92
Go = 33
Groovy = 7
HTML = 30
Hack = 20
Handlebars = 90
Haskell = 44
Idris = 46
Ini = 72
J = 51
JSON = 75
Java = 6
JavaScript = 22
JavaScriptReact = 93
Jsonnet = 76
Julia = 55
Kotlin = 4
LaTeX = 83
Lean = 48
Less = 27
Lua = 12
Makefile = 79
Markdown = 84
Matlab = 52
Nix = 77
OCaml = 41
Objective_C = 36
Objective_CPP = 37
PHP = 19
PLSQL = 70
Perl = 13
PowerShell = 67
Prolog = 71
Python = 15
R = 54
Racket = 11
Raku = 14
Razor = 62
ReST = 85
Ruby = 16
Rust = 40
SAS = 61
SCSS = 29
SML = 43
SQL = 69
Sass = 28
Scala = 5
Scheme = 10
ShellScript = 64
Skylark = 78
Swift = 2
TOML = 73
TeX = 82
TypeScript = 23
TypeScriptReact = 94
VisualBasic = 63
Vue = 25
Wolfram = 53
XML = 31
XSL = 32
YAML = 74
Zig = 38


_INDEX = DESCRIPTOR.message_types_by_name["Index"]
_METADATA = DESCRIPTOR.message_types_by_name["Metadata"]
_TOOLINFO = DESCRIPTOR.message_types_by_name["ToolInfo"]
_DOCUMENT = DESCRIPTOR.message_types_by_name["Document"]
_SYMBOL = DESCRIPTOR.message_types_by_name["Symbol"]
_PACKAGE = DESCRIPTOR.message_types_by_name["Package"]
_DESCRIPTOR = DESCRIPTOR.message_types_by_name["Descriptor"]
_SYMBOLINFORMATION = DESCRIPTOR.message_types_by_name["SymbolInformation"]
_RELATIONSHIP = DESCRIPTOR.message_types_by_name["Relationship"]
_OCCURRENCE = DESCRIPTOR.message_types_by_name["Occurrence"]
_DIAGNOSTIC = DESCRIPTOR.message_types_by_name["Diagnostic"]
_DESCRIPTOR_SUFFIX = _DESCRIPTOR.enum_types_by_name["Suffix"]
Index = _reflection.GeneratedProtocolMessageType(
    "Index",
    (_message.Message,),
    {
        "DESCRIPTOR": _INDEX,
        "__module__": "scip_pb2"
        # @@protoc_insertion_point(class_scope:scip.Index)
    },
)
_sym_db.RegisterMessage(Index)

Metadata = _reflection.GeneratedProtocolMessageType(
    "Metadata",
    (_message.Message,),
    {
        "DESCRIPTOR": _METADATA,
        "__module__": "scip_pb2"
        # @@protoc_insertion_point(class_scope:scip.Metadata)
    },
)
_sym_db.RegisterMessage(Metadata)

ToolInfo = _reflection.GeneratedProtocolMessageType(
    "ToolInfo",
    (_message.Message,),
    {
        "DESCRIPTOR": _TOOLINFO,
        "__module__": "scip_pb2"
        # @@protoc_insertion_point(class_scope:scip.ToolInfo)
    },
)
_sym_db.RegisterMessage(ToolInfo)

Document = _reflection.GeneratedProtocolMessageType(
    "Document",
    (_message.Message,),
    {
        "DESCRIPTOR": _DOCUMENT,
        "__module__": "scip_pb2"
        # @@protoc_insertion_point(class_scope:scip.Document)
    },
)
_sym_db.RegisterMessage(Document)

Symbol = _reflection.GeneratedProtocolMessageType(
    "Symbol",
    (_message.Message,),
    {
        "DESCRIPTOR": _SYMBOL,
        "__module__": "scip_pb2"
        # @@protoc_insertion_point(class_scope:scip.Symbol)
    },
)
_sym_db.RegisterMessage(Symbol)

Package = _reflection.GeneratedProtocolMessageType(
    "Package",
    (_message.Message,),
    {
        "DESCRIPTOR": _PACKAGE,
        "__module__": "scip_pb2"
        # @@protoc_insertion_point(class_scope:scip.Package)
    },
)
_sym_db.RegisterMessage(Package)

Descriptor = _reflection.GeneratedProtocolMessageType(
    "Descriptor",
    (_message.Message,),
    {
        "DESCRIPTOR": _DESCRIPTOR,
        "__module__": "scip_pb2"
        # @@protoc_insertion_point(class_scope:scip.Descriptor)
    },
)
_sym_db.RegisterMessage(Descriptor)

SymbolInformation = _reflection.GeneratedProtocolMessageType(
    "SymbolInformation",
    (_message.Message,),
    {
        "DESCRIPTOR": _SYMBOLINFORMATION,
        "__module__": "scip_pb2"
        # @@protoc_insertion_point(class_scope:scip.SymbolInformation)
    },
)
_sym_db.RegisterMessage(SymbolInformation)

Relationship = _reflection.GeneratedProtocolMessageType(
    "Relationship",
    (_message.Message,),
    {
        "DESCRIPTOR": _RELATIONSHIP,
        "__module__": "scip_pb2"
        # @@protoc_insertion_point(class_scope:scip.Relationship)
    },
)
_sym_db.RegisterMessage(Relationship)

Occurrence = _reflection.GeneratedProtocolMessageType(
    "Occurrence",
    (_message.Message,),
    {
        "DESCRIPTOR": _OCCURRENCE,
        "__module__": "scip_pb2"
        # @@protoc_insertion_point(class_scope:scip.Occurrence)
    },
)
_sym_db.RegisterMessage(Occurrence)

Diagnostic = _reflection.GeneratedProtocolMessageType(
    "Diagnostic",
    (_message.Message,),
    {
        "DESCRIPTOR": _DIAGNOSTIC,
        "__module__": "scip_pb2"
        # @@protoc_insertion_point(class_scope:scip.Diagnostic)
    },
)
_sym_db.RegisterMessage(Diagnostic)

if _descriptor._USE_C_DESCRIPTORS is False:
    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = (
        b"Z-github.com/sourcegraph/scip/bindings/go/scip/"
    )
    _SYNTAXKIND._options = None
    _SYNTAXKIND._serialized_options = b"\020\001"
    _SYNTAXKIND.values_by_name["IdentifierKeyword"]._options = None
    _SYNTAXKIND.values_by_name[
        "IdentifierKeyword"
    ]._serialized_options = b"\010\001"
    _SYNTAXKIND.values_by_name["IdentifierModule"]._options = None
    _SYNTAXKIND.values_by_name[
        "IdentifierModule"
    ]._serialized_options = b"\010\001"
    _DESCRIPTOR_SUFFIX._options = None
    _DESCRIPTOR_SUFFIX._serialized_options = b"\020\001"
    _DESCRIPTOR_SUFFIX.values_by_name["Package"]._options = None
    _DESCRIPTOR_SUFFIX.values_by_name[
        "Package"
    ]._serialized_options = b"\010\001"
    _PROTOCOLVERSION._serialized_start = 1495
    _PROTOCOLVERSION._serialized_end = 1544
    _TEXTENCODING._serialized_start = 1546
    _TEXTENCODING._serialized_end = 1610
    _SYMBOLROLE._serialized_start = 1612
    _SYMBOLROLE._serialized_end = 1737
    _SYNTAXKIND._serialized_start = 1740
    _SYNTAXKIND._serialized_end = 2614
    _SEVERITY._serialized_start = 2616
    _SEVERITY._serialized_end = 2702
    _DIAGNOSTICTAG._serialized_start = 2704
    _DIAGNOSTICTAG._serialized_end = 2782
    _LANGUAGE._serialized_start = 2785
    _LANGUAGE._serialized_end = 3905
    _INDEX._serialized_start = 20
    _INDEX._serialized_end = 147
    _METADATA._serialized_start = 150
    _METADATA._serialized_end = 309
    _TOOLINFO._serialized_start = 311
    _TOOLINFO._serialized_end = 371
    _DOCUMENT._serialized_start = 374
    _DOCUMENT._serialized_end = 506
    _SYMBOL._serialized_start = 508
    _SYMBOL._serialized_end = 603
    _PACKAGE._serialized_start = 605
    _PACKAGE._serialized_end = 662
    _DESCRIPTOR._serialized_start = 665
    _DESCRIPTOR._serialized_end = 923
    _DESCRIPTOR_SUFFIX._serialized_start = 758
    _DESCRIPTOR_SUFFIX._serialized_end = 923
    _SYMBOLINFORMATION._serialized_start = 925
    _SYMBOLINFORMATION._serialized_end = 1026
    _RELATIONSHIP._serialized_start = 1029
    _RELATIONSHIP._serialized_end = 1159
    _OCCURRENCE._serialized_start = 1162
    _OCCURRENCE._serialized_end = 1362
    _DIAGNOSTIC._serialized_start = 1365
    _DIAGNOSTIC._serialized_end = 1493
# @@protoc_insertion_point(module_scope)



"""
Abstract base class for completion pipelines.
"""
import logging
import uuid
from abc import abstractmethod
from typing import Any, Optional, Union

from openai.types import Completion
from openai.types.chat import ChatCompletion

from ..providers.llm import GenerationConfig, LLMProvider
from ..providers.logging import LoggingDatabaseConnection, log_execution_to_db
from .pipeline import Pipeline

logger = logging.getLogger(__name__)

DEFAULT_SYSTEM_PROMPT = "You are a helpful assistant."
DEFAULT_TASK_PROMPT = """
## Task:
Answer the query given immediately below given the context which follows later.

### Query:
{query}

### Context:
{context}

### Query:
{query}

## Response:
"""


class RAGPipeline(Pipeline):
    def __init__(
        self,
        llm: "LLMProvider",
        generation_config: "GenerationConfig",
        system_prompt: Optional[str] = None,
        task_prompt: Optional[str] = None,
        logging_provider: Optional[LoggingDatabaseConnection] = None,
        **kwargs,
    ):
        self.llm = llm
        self.generation_config = generation_config
        self.system_prompt = system_prompt or DEFAULT_SYSTEM_PROMPT
        self.task_prompt = task_prompt or DEFAULT_TASK_PROMPT
        self.logging_provider = logging_provider
        self.pipeline_run_info = None
        super().__init__(logging_provider=logging_provider, **kwargs)

    def initialize_pipeline(self, query: str, search_only: bool) -> None:
        self.pipeline_run_info = {
            "run_id": uuid.uuid4(),
            "type": "rag" if not search_only else "search",
        }
        self.ingress(query)

    @log_execution_to_db
    def ingress(self, data: Any) -> Any:
        """
        Ingresses data into the pipeline.
        """
        self._check_pipeline_initialized()
        return data

    @abstractmethod
    def transform_query(self, query: str) -> Any:
        """
        Transforms the input query for retrieval.
        """
        pass

    @abstractmethod
    def search(
        self,
        transformed_query,
        filters: dict[str, Any],
        limit: int,
        *args,
        **kwargs,
    ) -> list:
        """
        Retrieves results based on the transformed query.
        The search_type parameter allows for specifying the type of search,
        """
        pass

    @abstractmethod
    def rerank_results(self, results: list) -> list:
        """
        Reranks the retrieved results based on relevance or other criteria.
        """
        pass

    @abstractmethod
    def _get_extra_args(self, *args, **kwargs) -> dict[str, Any]:
        """
        Returns extra arguments for the generation request.
        """
        pass

    @abstractmethod
    def _format_results(self, results: list) -> str:
        """
        Formats the results for generation.
        """
        pass

    @log_execution_to_db
    def construct_context(
        self,
        results: list,
    ) -> str:
        reranked_results = self.rerank_results(results)
        return self._format_results(reranked_results)

    @log_execution_to_db
    def construct_prompt(self, inputs: dict[str, str]) -> str:
        """
        Constructs a prompt for generation based on the reranked chunks.
        """
        return self.task_prompt.format(**inputs)

    @log_execution_to_db
    def generate_completion(
        self,
        prompt: str,
        generate_with_chat=True,
    ) -> Union[ChatCompletion, Completion]:
        """
        Generates a completion based on the prompt.
        """
        self._check_pipeline_initialized()
        if generate_with_chat:
            return self.llm.get_chat_completion(
                [
                    {
                        "role": "system",
                        "content": self.system_prompt,
                    },
                    {
                        "role": "user",
                        "content": prompt,
                    },
                ],
                self.generation_config,
                **self._get_extra_args(),
            )
        else:
            raise NotImplementedError(
                "Generation without chat is not implemented yet."
            )

    def run(
        self, query, filters={}, limit=10, search_only=False
    ) -> Union[ChatCompletion, Completion, list]:
        """
        Runs the completion pipeline.
        """
        self.initialize_pipeline(query, search_only)

        logger.debug(f"Pipeline run type: {self.pipeline_run_info}")

        transformed_query = self.transform_query(query)
        search_results = self.search(transformed_query, filters, limit)
        if search_only:
            logger.debug(f"Pipeline run type: {self.pipeline_run_info}")
            return search_results

        context = self.construct_context(search_results)
        prompt = self.construct_prompt(
            {"query": transformed_query, "context": context}
        )
        return self.generate_completion(prompt, generate_with_chat=True)



import glob
import uuid

import fire

from r2r.client import R2RClient


class PDFChat:
    def __init__(self, base_url="http://localhost:8000", user_id=None):
        self.client = R2RClient(base_url)
        if not user_id:
            self.user_id = str(uuid.uuid5(uuid.NAMESPACE_DNS, "user_id"))
        self.titles = {
            "examples/pdf_chat/meditations.pdf": "Title: Meditations - Marcus Aurelius",
            # uncomment the following line to add more documents
            # "examples/pdf_chat/the_republic.pdf": "Title: The Republic - Plato",
        }

    def ingest(self):
        for file_path in glob.glob("examples/pdf_chat/*.pdf"):
            if file_path in self.titles:
                print("Uploading file: ", file_path)
                document_id = str(uuid.uuid5(uuid.NAMESPACE_DNS, file_path))
                metadata = {
                    "user_id": self.user_id,
                    "chunk_prefix": self.titles[file_path],
                }
                settings = {}
                upload_response = self.client.upload_and_process_file(
                    document_id, file_path, metadata, settings
                )
                print("Upload response = ", upload_response)

    def search(self, query):
        search_response = self.client.search(
            query,
            5,
            filters={"user_id": self.user_id},
        )
        for i, response in enumerate(search_response):
            text = response["metadata"]["text"]
            title, body = text.split("\n", 1)
            print(f"Result {i + 1}: {title}")
            print(body[:500])
            print("\n")

    def rag_completion(self, query):
        rag_response = self.client.rag_completion(
            query,
            5,
            filters={"user_id": self.user_id},
        )
        print("rag_response = ", rag_response)

    def delete_document(self, document_path: str):
        document_id = str(uuid.uuid5(uuid.NAMESPACE_DNS, document_path))
        response = self.client.filtered_deletion("document_id", document_id)
        print("Deletion response = ", response)


if __name__ == "__main__":
    fire.Fire(PDFChat)



"""A simple example to demonstrate the usage of `BasicRAGPipeline`."""
import logging

import dotenv

from r2r.core import GenerationConfig, LoggingDatabaseConnection
from r2r.embeddings import OpenAIEmbeddingProvider
from r2r.llms import OpenAIConfig, OpenAILLM
from r2r.main import load_config
from r2r.pipelines import BasicRAGPipeline
from r2r.vector_dbs import PGVectorDB, QdrantDB


class DemoRAGPipeline(BasicRAGPipeline):
    # Modifies `BasicRAGPipeline` run to return search_results and completion
    def run(self, query, filters={}, limit=10, search_only=False):
        """
        Runs the completion pipeline.
        """
        self.initialize_pipeline(query, search_only)
        transformed_query = self.transform_query(query)
        search_results = self.search(transformed_query, filters, limit)
        if search_only:
            return search_results, None
        context = self.construct_context(search_results)
        prompt = self.construct_prompt(
            {"query": transformed_query, "context": context}
        )
        completion = self.generate_completion(prompt)
        return search_results, completion


if __name__ == "__main__":
    dotenv.load_dotenv()

    (
        api_config,
        logging_config,
        embedding_config,
        database_config,
        llm_config,
        text_splitter_config,
    ) = load_config()

    query = "Is the answer to the question `What are the energy levels for a particle in a box?` contained in the search results shown below?"

    logger = logging.getLogger(logging_config["name"])
    logging.basicConfig(level=logging_config["level"])

    logger.debug("Starting the rag pipeline.")

    embeddings_provider = OpenAIEmbeddingProvider()
    embedding_model = embedding_config["model"]
    embedding_dimension = embedding_config["dimension"]
    embedding_batch_size = embedding_config["batch_size"]

    db = (
        QdrantDB() if database_config["provider"] == "qdrant" else PGVectorDB()
    )
    collection_name = database_config["collection_name"]
    db.initialize_collection(collection_name, embedding_dimension)

    llm = OpenAILLM(OpenAIConfig())
    generation_config = GenerationConfig(
        model_name=llm_config["model_name"],
        temperature=llm_config["temperature"],
        top_p=llm_config["top_p"],
        top_k=llm_config["top_k"],
        max_tokens_to_sample=llm_config["max_tokens_to_sample"],
        do_stream=llm_config["do_stream"],
    )

    logging_provider = LoggingDatabaseConnection(logging_config["database"])
    pipeline = DemoRAGPipeline(
        llm,
        generation_config,
        db=db,
        logging_provider=logging_provider,
        embedding_model=embedding_model,
        embeddings_provider=embeddings_provider,
    )

    search_results, completion_1 = pipeline.run(query, search_only=False)

    for result in search_results:
        logger.info("-" * 100)
        logger.info(f"Search Result:\n{result}")

    # To delete the primary Physics document from the collection
    db.filtered_deletion("document_id", "a9b92938-12e6-5ea4-b412-a4a1d4b48a0c")

    search_results, completion_2 = pipeline.run(query, search_only=False)

    logger.info("After Deletion: ")
    for result in search_results:
        logger.info("-" * 100)
        logger.info(f"Search Result:\n{result}")

    logger.info("-" * 100)
    logger.info(f"Completion 1 Result:\n{completion_1}")
    logger.info(f"Completion 2 Result:\n{completion_2}")
    logger.info("-" * 100)

    pipeline.close()



import logging

import dotenv
from langchain.text_splitter import RecursiveCharacterTextSplitter

from r2r.core import GenerationConfig, LoggingDatabaseConnection
from r2r.llms import OpenAIConfig, OpenAILLM
from r2r.pipelines import (
    BasicEmbeddingPipeline,
    BasicIngestionPipeline,
    BasicRAGPipeline,
)

from .app import create_app
from .utils import load_config

dotenv.load_dotenv()


class E2EPipelineFactory:
    @staticmethod
    def get_db(database_config):
        if database_config["provider"] == "qdrant":
            from r2r.vector_dbs import QdrantDB

            return QdrantDB()
        elif database_config["provider"] == "pgvector":
            from r2r.vector_dbs import PGVectorDB

            return PGVectorDB()
        elif database_config["provider"] == "local":
            from r2r.vector_dbs import LocalVectorDB

            return LocalVectorDB()

    @staticmethod
    def get_embeddings_provider(embedding_config: dict):
        if embedding_config["provider"] == "openai":
            from r2r.embeddings import OpenAIEmbeddingProvider

            return OpenAIEmbeddingProvider()
        elif embedding_config["provider"] == "sentence_transformers":
            from r2r.embeddings import SentenceTransformerEmbeddingProvider

            return SentenceTransformerEmbeddingProvider()

    @staticmethod
    def get_llm():
        return OpenAILLM(OpenAIConfig())

    @staticmethod
    def get_text_splitter(text_splitter_config):
        return RecursiveCharacterTextSplitter(
            chunk_size=text_splitter_config["chunk_size"],
            chunk_overlap=text_splitter_config["chunk_overlap"],
            length_function=len,
            is_separator_regex=False,
        )

    @staticmethod
    def create_pipeline(
        db=None,
        embeddings_provider=None,
        llm=None,
        text_splitter=None,
        generation_config=None,
        ingestion_pipeline_impl=BasicIngestionPipeline,
        embedding_pipeline_impl=BasicEmbeddingPipeline,
        rag_pipeline_impl=BasicRAGPipeline,
        app_fn=create_app,
        config_path=None,
    ):
        (
            api_config,
            logging_config,
            embedding_config,
            database_config,
            llm_config,
            text_splitter_config,
        ) = load_config(config_path)

        logging.basicConfig(level=logging_config["level"])

        embeddings_provider = (
            embeddings_provider
            or E2EPipelineFactory.get_embeddings_provider(embedding_config)
        )
        # TODO - Encapsulate the embedding metadata into a container
        embedding_model = embedding_config["model"]
        embedding_dimension = embedding_config["dimension"]
        embedding_batch_size = embedding_config["batch_size"]

        db = db or E2EPipelineFactory.get_db(database_config)
        collection_name = database_config["collection_name"]
        db.initialize_collection(collection_name, embedding_dimension)

        llm = llm or E2EPipelineFactory.get_llm()
        generation_config = generation_config or GenerationConfig(
            model_name=llm_config["model_name"],
            temperature=llm_config["temperature"],
            top_p=llm_config["top_p"],
            top_k=llm_config["top_k"],
            max_tokens_to_sample=llm_config["max_tokens_to_sample"],
            do_stream=llm_config["do_stream"],
        )

        logging_provider = LoggingDatabaseConnection(
            logging_config["provider"], logging_config["database"]
        )

        cmpl_pipeline = rag_pipeline_impl(
            llm,
            generation_config,
            db=db,
            embedding_model=embedding_model,
            embeddings_provider=embeddings_provider,
            logging_provider=logging_provider,
        )

        text_splitter = text_splitter or E2EPipelineFactory.get_text_splitter(
            text_splitter_config
        )

        embd_pipeline = embedding_pipeline_impl(
            embedding_model,
            embeddings_provider,
            db,
            logging_provider=logging_provider,
            text_splitter=text_splitter,
            embedding_batch_size=embedding_batch_size,
        )
        # TODO - Set ingestion class in config file
        ingst_pipeline = ingestion_pipeline_impl()

        app = app_fn(
            ingestion_pipeline=ingst_pipeline,
            embedding_pipeline=embd_pipeline,
            rag_pipeline=cmpl_pipeline,
            logging_provider=logging_provider,
        )

        return app



"""
A simple example to demonstrate the usage of `BasicIngestionPipeline`.
"""
import collections
import copy
import json
import logging
from enum import Enum
from typing import Optional, Union

from r2r.core import IngestionPipeline, LoggingDatabaseConnection

logger = logging.getLogger(__name__)


class EntryType(Enum):
    TXT = "txt"
    JSON = "json"
    HTML = "html"
    PDF = "pdf"


class BasicIngestionPipeline(IngestionPipeline):
    """
    Processes incoming documents into plaintext based on their data type.
    Supports TXT, JSON, HTML, and PDF formats.
    """

    def __init__(
        self,
        logging_provider: Optional[LoggingDatabaseConnection] = None,
    ):
        try:
            from bs4 import BeautifulSoup

            self.BeautifulSoup = BeautifulSoup
        except ImportError:
            raise ValueError(
                "Error, `bs4` is requried to run `BasicIngestionPipeline`. Please install it using `pip install bs4`, or if using poetry in local development, `poetry install -E parsing`."
            )

        try:
            from pypdf import PdfReader

            self.PdfReader = PdfReader
        except ImportError:
            raise ValueError(
                "Error, `pypdf` is requried to run `BasicIngestionPipeline`. Please install it using `pip install pypdf` or if using poetry in local development, `poetry install -E parsing`."
            )

        logger.info(
            f"Initalizing a `BasicIngestionPipeline` to process incoming documents."
        )

        super().__init__(
            logging_provider,
        )
        self.pipeline_run_info = None

    @property
    def supported_types(self) -> list[str]:
        """
        Lists the data types supported by the pipeline.
        """
        return [entry_type.value for entry_type in EntryType]

    def process_data(
        self,
        entry_type: str,
        entry_data: Union[bytes, str],
    ) -> str:
        """
        Process data into plaintext based on the data type.
        """
        if entry_type == EntryType.TXT.value:
            if not isinstance(entry_data, str):
                raise ValueError("TXT data must be a string.")
            return entry_data
        elif entry_type == EntryType.JSON.value:
            try:
                entry_json = json.loads(entry_data)
            except json.JSONDecodeError:
                raise ValueError("JSON data must be a valid JSON string.")
            return self._parse_json(entry_json)
        elif entry_type == EntryType.HTML.value:
            if not isinstance(entry_data, str):
                raise ValueError("HTML data must be a string.")
            return self._parse_html(entry_data)
        elif entry_type == EntryType.PDF.value:
            if not isinstance(entry_data, bytes):
                raise ValueError("PDF data must be a bytes object.")
            return self._parse_pdf(entry_data)
        else:
            raise ValueError(f"EntryType {entry_type} not supported.")

    def parse_entry(
        self, entry_type: str, entry_data: Union[bytes, str]
    ) -> str:
        """
        Parse entry data into plaintext based on the entry type.
        """
        return self.process_data(entry_type, entry_data)

    def _parse_json(self, data: dict) -> str:
        """
        Parse JSON data into plaintext.
        """

        def remove_objects_with_null(obj):
            if not isinstance(obj, collections.abc.Mapping):
                return obj
            result = copy.deepcopy(obj)
            for key, value in obj.items():
                if isinstance(value, collections.abc.Mapping):
                    result[key] = remove_objects_with_null(value)
                elif value is None:
                    del result[key]
            return result

        def format_json_as_text(obj, indent=0):
            """
            Recursively formats a JSON object as a text document, placing nested components on new lines with indentation.

            Args:
                obj: The JSON object to format.
                indent: The current indentation level.

            Returns:
                A string representing the formatted text document.
            """
            lines = []
            indent_str = "" * indent  # Two spaces per indentation level

            if isinstance(obj, dict):
                for key, value in obj.items():
                    if isinstance(value, (dict, list)):
                        nested = format_json_as_text(value, indent + 1)
                        lines.append(f"{indent_str}{key}:\n{nested}")
                    else:
                        lines.append(f"{indent_str}{key}: {value}")
            elif isinstance(obj, list):
                for item in obj:
                    nested = format_json_as_text(item, indent + 1)
                    lines.append(f"{nested}")
            else:
                return f"{indent_str}{obj}"

            return "\n".join(lines)

        return format_json_as_text(remove_objects_with_null(data))

    def _parse_html(self, data: str) -> str:
        """
        Parse HTML data into plaintext.
        """

        soup = self.BeautifulSoup(data, "html.parser")
        return soup.get_text()

    def _parse_pdf(self, file_data: bytes) -> str:
        import string
        from io import BytesIO

        """
        Process PDF file data into plaintext.
        """
        pdf = self.PdfReader(BytesIO(file_data))
        text = ""
        for page in pdf.pages:
            page_text = page.extract_text()
            if page_text is not None:
                # Remove non-printable characters
                page_text = "".join(
                    filter(lambda x: x in string.printable, page_text)
                )
                text += page_text + "\n"
        return text



import logging
import os
from typing import Generator, Optional, Tuple

from r2r.core import DatasetConfig, DatasetProvider

logger = logging.getLogger(__name__)


class PdfDataReader(DatasetProvider):
    def __init__(self, provider: str = "pdf"):
        if provider != "pdf":
            raise ValueError(
                f"Error, PdfDataReader was passed `{provider}` as a provider."
            )
        super().__init__(provider)
        try:
            from pypdf import PdfReader

            self.PdfReader = PdfReader
            self.pdfs: list[Tuple[PdfReader, DatasetConfig]] = []
        except ImportError:
            raise ValueError(
                "Error, `pypdf` is requried to run `PdfDataReader`. Please install it using `pip install pypdf` or if using poetry in local development, `poetry install -E parsing`."
            )

    def load_datasets(
        self,
        dataset_configs: list[DatasetConfig],
        split: str = "train",
        streaming: bool = True,
    ) -> None:
        logger.info("Loading datasets with HuggingFaceDataProvider now.")
        for config in dataset_configs:
            logger.info(
                f"Loading dataset {config.name} with files {config.data_files}."
            )
            if config.name is None:
                raise ValueError(
                    f"Error, the dataset name must be specified in the `DatasetConfig`."
                )
            if len(config.data_files) == 0:
                raise ValueError(
                    f"Error, the dataset files must be specified in the `DatasetConfig`."
                )
            for file_path in config.data_files:
                if os.path.isfile(file_path):
                    if ".pdf" not in file_path:
                        raise ValueError(
                            f"Error, the file {file_path} is not a pdf file."
                        )
                    pdf = self.PdfReader(file_path)
                    self.pdfs.append((pdf, config))
                elif os.path.isdir(file_path):
                    for root, dirs, files in os.walk(file_path):
                        for file in files:
                            if ".pdf" in file:
                                pdf = self.PdfReader(os.path.join(root, file))
                                self.pdfs.append((pdf, config))

    def stream_text(
        self,
    ) -> Generator[Optional[Tuple[str, DatasetConfig]], None, None]:
        logger.info("Streaming text with with PdfDataReader now.")
        if self.pdfs == []:
            raise ValueError(
                f"Error, no pdfs were loaded with `PdfDataReader`."
            )
        for pdf, config in self.pdfs:
            entries_streamed = 0
            for page in pdf.pages:
                if (
                    config.max_entries is not None
                    and entries_streamed >= config.max_entries
                ):
                    break  # Stop streaming if max_entries is reached

                text = page.extract_text()

                if text:
                    yield text, config
                    entries_streamed += 1



from r2r.main import E2EPipelineFactory

# Creates a pipeline with default configuration
# This is the main entry point for the application
# The pipeline is built using the `config.json` file
app = E2EPipelineFactory.create_pipeline()



from .abstractions.document import BasicDocument
from .pipelines.embedding import EmbeddingPipeline
from .pipelines.ingestion import IngestionPipeline
from .pipelines.rag import RAGPipeline
from .providers.dataset import DatasetConfig, DatasetProvider
from .providers.embedding import EmbeddingProvider
from .providers.llm import GenerationConfig, LLMConfig, LLMProvider
from .providers.logging import LoggingDatabaseConnection, log_execution_to_db
from .providers.vector_db import (
    VectorDBProvider,
    VectorEntry,
    VectorSearchResult,
)

__all__ = [
    "BasicDocument",
    "EmbeddingPipeline",
    "IngestionPipeline",
    "RAGPipeline",
    "LoggingDatabaseConnection",
    "log_execution_to_db",
    "DatasetConfig",
    "DatasetProvider",
    "EmbeddingProvider",
    "GenerationConfig",
    "LLMConfig",
    "LLMProvider",
    "VectorSearchResult",
    "VectorEntry",
    "VectorDBProvider",
]



import glob
import uuid

from r2r.client import R2RClient

# Initialize the client with the base URL of your API
base_url = "http://localhost:8000"  # Change this to your actual API base URL
client = R2RClient(base_url)

titles = {
    "examples/academy/meditations.pdf": "Title: Meditations - Marcus Aurelius",
    "examples/academy/the_republic.pdf": "Title: The Republic - Plato",
}

user_id_0 = str(uuid.uuid5(uuid.NAMESPACE_DNS, "user_0"))

for file_path in glob.glob("examples/academy/*.pdf"):
    print(f"Uploading and processing file: {file_path}")
    # # Upload and process a file
    document_id = str(uuid.uuid5(uuid.NAMESPACE_DNS, file_path))
    metadata = {"user_id": user_id_0, "chunk_prefix": titles[file_path]}
    settings = {}
    upload_response = client.upload_and_process_file(
        document_id, file_path, metadata, settings
    )

prompt = """You are given a user query {query} and a user context {context}. Use the context to answer the given query. """
formatted_prompt = prompt.format(
    query="What are the key themes of these books?",
    context="User Uploads:\nTitle: Meditations - Marcus Aurelius\nTitle: The Republic - Plato",
)
# Perform a search on this file
search_response = client.rag_completion(
    formatted_prompt, 5, filters={"user_id": user_id_0}
)
print(search_response)



"""A simple example to demonstrate the usage of `WebSearchRAGPipeline`."""
from r2r.main import E2EPipelineFactory
from r2r.pipelines import WebSearchRAGPipeline

# Creates a pipeline using the `WebSearchRAGPipeline` implementation
app = E2EPipelineFactory.create_pipeline(
    rag_pipeline_impl=WebSearchRAGPipeline
)
